package be.iminds.ilabt.jfed.ui.commandline;

import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Server;
import be.iminds.ilabt.jfed.lowlevel.GeniUser;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.SimpleGeniUser;
import be.iminds.ilabt.jfed.lowlevel.TestbedInfoSource;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityFinder;

import be.iminds.ilabt.jfed.util.GeniUrn;
import com.google.inject.AbstractModule;
import com.google.inject.Provides;
import com.google.inject.name.Names;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.inject.Named;
import java.io.File;
import java.util.ArrayList;
import java.util.List;

/**
 * BaseCliModule
 */
public class BaseCliKeyCertUserModule extends AbstractModule {
    private final static Logger LOG = LoggerFactory.getLogger(BaseCliKeyCertUserModule.class);

    private final String keyCertContent;
    private final char[] keyPass;
    private final File certificateKeyFile;
    private final File privateKeyFile;

    public BaseCliKeyCertUserModule(String keyCertContent, char[] keyPass, File certificateKeyFile, File privateKeyFile) {
        this.keyCertContent = keyCertContent;
        this.keyPass = keyPass;
        this.certificateKeyFile = certificateKeyFile;
        this.privateKeyFile = privateKeyFile;
    }

    @Override
    protected void configure() {
        SimpleGeniUser geniUser = new SimpleGeniUser(null, null, keyCertContent, keyPass, certificateKeyFile, privateKeyFile);
        geniUser.deriveUserAndAuthFromPemContent(null);
        assert geniUser.getUserUrn() != null;
        bind(GeniUrn.class).annotatedWith(Names.named("userUrn")).toInstance(geniUser.getUserUrn());
    }

    @Provides @Named("userServer")
    public Server provideGeniUserServer(@Named("userUrn") GeniUrn userUrn, TestbedInfoSource testbedInfoSource, AuthorityFinder authorityFinder) throws BaseCli.CliArgumentException {
        assert testbedInfoSource != null;
        assert authorityFinder != null;
        assert userUrn != null;

        SimpleGeniUser geniUser = new SimpleGeniUser(null, userUrn, keyCertContent, keyPass, certificateKeyFile, privateKeyFile);
        geniUser.deriveUserAndAuthFromPemContent(authorityFinder);
        if (geniUser.getUserAuthorityServer() == null)
            if (authorityFinder.size() < 10) {
                List<String> authUrnStrings = new ArrayList<String>();
                for (Server auth : testbedInfoSource.getServers())
                    authUrnStrings.add(auth.getDefaultComponentManagerUrn());
                throw new BaseCli.CliArgumentException("User Authority not found: \"" + userUrn + "\". authorityFinder knows " + authorityFinder.size() + " authorities: "+authUrnStrings);
            }
            else
                throw new BaseCli.CliArgumentException("User Authority not found: \"" + userUrn + "\". authorityFinder knows " + authorityFinder.size() + " authorities");

        return geniUser.getUserAuthorityServer();
    }

    @Provides
    public GeniUser provideGeniUser(@Named("userServer") Server userServer, @Named("userUrn") GeniUrn userUrn) {
        assert userUrn != null;
        assert userServer != null;
        return new SimpleGeniUser(userServer, userUrn, keyCertContent, keyPass, certificateKeyFile, privateKeyFile);
    }

    @Provides
    public GeniUserProvider provideGeniUserProvider(final GeniUser geniUser) {
        return new GeniUserProvider() {
            @Override
            public GeniUser getLoggedInGeniUser() {
                return geniUser;
            }

            @Override
            public boolean isUserLoggedIn() {
                return geniUser != null;
            }
        };
    }
}
