package be.iminds.ilabt.jfed.ui.javafx.choosers;

import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Server;
import be.iminds.ilabt.jfed.lowlevel.TestbedInfoSource;
import be.iminds.ilabt.jfed.ui.javafx.dialogs.JFDialogs;
import com.cathive.fx.guice.FXMLComponent;
import javafx.beans.property.BooleanProperty;
import javafx.beans.property.ReadOnlyObjectProperty;
import javafx.beans.property.SimpleBooleanProperty;
import javafx.collections.FXCollections;
import javafx.collections.ObservableList;
import javafx.fxml.FXML;
import javafx.scene.control.Button;
import javafx.scene.control.ComboBox;
import javafx.scene.layout.VBox;
import javafx.util.StringConverter;

import javax.inject.Inject;
import javax.inject.Provider;

@FXMLComponent
public class AuthorityChooser extends VBox {
    @FXML
    private ComboBox<Server> comboBox;
    @FXML
    private Button editButton;

    private final TestbedInfoSource testbedInfoSource;

    @Inject
    public AuthorityChooser(TestbedInfoSource testbedInfoSource) {
        this.testbedInfoSource = testbedInfoSource;
    }

    @FXML
    private void initialize() {

        comboBox.setConverter(new StringConverter<Server>() {
            @Override
            public String toString(Server server) {
                assert server != null;
                return server.getName();
            }

            @Override
            public Server fromString(String s) {
                throw new RuntimeException("unused");
            }
        });

        editable.addListener((observableValue, oldBoolean, newBoolean) -> comboBox.disableProperty().set(!newBoolean));

        showEditListButton.addListener((observableValue, oldBoolean, newBoolean) -> {
            editButton.setVisible(newBoolean);
            editButton.setManaged(newBoolean);
        });

        assert (testbedInfoSource != null);
        comboBox.setItems(FXCollections.observableArrayList(testbedInfoSource.getServers()));
        comboBox.getSelectionModel().selectFirst();
    }


    public Server getSelectedServer() {
        return comboBox.getSelectionModel().getSelectedItem();
    }

    public ReadOnlyObjectProperty<Server> selectedAuthorityProperty() {
        return comboBox.getSelectionModel().selectedItemProperty();
    }

    public void select(Server server) {
        assert server != null : "must always select something";
        comboBox.getSelectionModel().select(server);
    }

    @FXML
    public void editList() {
        JFDialogs.create().message("Editing the Server list is currently not supported").showError();
        //TODO: make this work again
//        amListProvider.get().show();
    }

    private BooleanProperty editable = new SimpleBooleanProperty(true);

    public BooleanProperty editableProperty() {
        return editable;
    }

    private BooleanProperty showEditListButton = new SimpleBooleanProperty(true);

    public BooleanProperty showEditListButtonProperty() {
        return showEditListButton;
    }

    private ObservableList<Server> filteredAuthorities = FXCollections.observableArrayList();
    private ServerFilter serverFilter;
//    private ListChangeListener<Server> listChangeListener;

    public void setServerFilter(final ServerFilter serverFilter) {
        assert testbedInfoSource.getServers() != null;

        if (serverFilter == null) {
            this.serverFilter = null;
            comboBox.setItems(FXCollections.observableArrayList(testbedInfoSource.getServers()));
            comboBox.getSelectionModel().selectFirst();
            return;
        }
        this.serverFilter = serverFilter;

        filteredAuthorities.clear();
        for (Server server : testbedInfoSource.getServers()) {
            if (serverFilter.includeAuthority(server))
                filteredAuthorities.add(server);
        }
        comboBox.setItems(filteredAuthorities);

        //init and set listener only once. It keeps listening, but it doesn't do anything when there is no filter
//        if (listChangeListener != null) {
//            listChangeListener = change -> {
//                if (AuthorityChooser.this.serverFilter == null)
//                    return;
//
//                while (change.next()) {
//                    if (change.wasAdded()) {
//                        for (Server a : change.getAddedSubList())
//                            if (AuthorityChooser.this.serverFilter.includeAuthority(a))
//                                filteredAuthorities.add(a);
//                    }
//                    if (change.wasRemoved())
//                        filteredAuthorities.removeAll(change.getRemoved());
//                }
//            };
//            testbedInfoSource.authorityInfosProperty().addListener(listChangeListener);
//        }
        comboBox.getSelectionModel().selectFirst();
    }

    public static interface ServerFilter {
        public boolean includeAuthority(Server server);
    }
}
