package be.iminds.ilabt.jfed.experimenter_gui.dialogs;

import be.iminds.ilabt.jfed.lowlevel.TestbedInfoSource;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityFinder;

import be.iminds.ilabt.jfed.fedmon.webapi.service.json.*;
import be.iminds.ilabt.jfed.ui.javafx.GlyphUtils;
import com.cathive.fx.guice.FXMLComponent;
import javafx.beans.property.BooleanProperty;
import javafx.beans.property.IntegerProperty;
import javafx.beans.property.SimpleBooleanProperty;
import javafx.beans.property.SimpleIntegerProperty;
import javafx.collections.FXCollections;
import javafx.collections.ObservableList;
import javafx.fxml.FXML;
import javafx.scene.Node;
import javafx.scene.control.ButtonType;
import javafx.scene.control.Dialog;
import javafx.scene.control.DialogPane;
import javafx.scene.control.ListView;
import javafx.scene.control.cell.CheckBoxListCell;
import javafx.scene.layout.VBox;
import javafx.scene.paint.Color;
import javafx.util.StringConverter;
import org.controlsfx.glyphfont.FontAwesome;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.inject.Inject;
import java.util.Comparator;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;

/**
 * User: twalcari
 * Date: 12/5/13
 * Time: 9:21 PM
 */
@FXMLComponent
public class SelectAuthoritiesDialog extends VBox {
    private static final Logger LOG = LoggerFactory.getLogger(SelectAuthoritiesDialog.class);

    private final ObservableList<SelectableAuthority> selectableAuthorities = FXCollections.observableArrayList();

    private final IntegerProperty selectedAuthoritiesCount = new SimpleIntegerProperty();

    @FXML
    private ListView<SelectableAuthority> authoritiesListView;
    private Dialog<Set<Server>> dialog;

    @Inject
    private TestbedInfoSource testbedInfoSource;
    @Inject
    private AuthorityFinder authorityFinder;

    SelectAuthoritiesDialog() {
    }

    @FXML
    private void initialize() {
        selectableAuthorities.addAll(
                testbedInfoSource.getServers().stream()
                        .filter(s -> !s.isEdgeVlan())
                        .map(SelectableAuthority::new)
                        .sorted(Comparator.comparing(auth -> auth.getExactSfaAuthority().getName().toLowerCase()))
                        .collect(Collectors.toList()));

        authoritiesListView.setItems(selectableAuthorities);


        authoritiesListView.setCellFactory(
                CheckBoxListCell.forListView(SelectableAuthority::selectedProperty,
                        new StringConverter<SelectableAuthority>() {
                            @Override
                            public String toString(SelectableAuthority selectableAuthority) {
                                return selectableAuthority.getExactSfaAuthority().getName();
                            }

                            @Override
                            public SelectableAuthority fromString(String s) {
                                return null;
                            }
                        }));

        DialogPane dialogPane = new DialogPane();
        dialogPane.setContent(this);
        dialogPane.setHeaderText("Select the involved authorities");
        dialogPane.setGraphic(GlyphUtils.createDialogGlyph(FontAwesome.Glyph.AMBULANCE, Color.GREEN));
        dialogPane.getButtonTypes().setAll(ButtonType.OK, ButtonType.CANCEL);

        Node okButton = dialogPane.lookupButton(ButtonType.OK);

        okButton.disableProperty().bind(selectedAuthoritiesCount.isEqualTo(0));


        dialog = new Dialog<>();
        dialog.setDialogPane(dialogPane);
        dialog.setResizable(false);

        dialog.setResultConverter(buttonType -> {
            if (buttonType == ButtonType.CANCEL)
                return null;
            else {
                return selectableAuthorities.stream()
                        .filter(SelectableAuthority::getSelected)
//                        .map(SelectableAuthority::getUsedSfaAuthority) caller has to worry about which auth to use, not us!
                        .map(SelectableAuthority::getExactSfaAuthority)
                        .collect(Collectors.toSet());
            }
        });
    }

    public Optional<Set<Server>> showDialog() {
        return dialog.showAndWait();
    }

    private class SelectableAuthority {
        private final BooleanProperty selected = new SimpleBooleanProperty(false);
        private final Server sfaAuthority;

        private SelectableAuthority(Server sfaAuthority) {
            this.sfaAuthority = sfaAuthority;

            selected.addListener((observable, oldValue, newValue) -> {
                if (newValue != oldValue) {
                    if (newValue) {
                        selectedAuthoritiesCount.set(selectedAuthoritiesCount.get() + 1);
                    } else {
                        selectedAuthoritiesCount.set(selectedAuthoritiesCount.get() - 1);
                    }
                }
            });

        }

        public boolean getSelected() {
            return selected.get();
        }

        public BooleanProperty selectedProperty() {
            return selected;
        }

        public void setSelected(boolean selected) {
            this.selected.set(selected);
        }

        public Server getExactSfaAuthority() {
            return sfaAuthority;
        }

        //caller has to worry about which auth to use, not us!
//        public Server getUsedSfaAuthority() {
//            assert authorityFinder != null;
//            assert sfaAuthority != null;
//            assert sfaAuthority.getDefaultComponentManagerAsGeniUrn() != null;
//            Server res = authorityFinder.findByUrn(sfaAuthority.getDefaultComponentManagerAsGeniUrn(), AuthorityFinder.Purpose.RECOVER);
//            if (res == null)
//                LOG.warn("Could not find authority for recover: \"" + sfaAuthority.getDefaultComponentManagerAsGeniUrn() + "\"");
//            return res;
//        }
    }
}
