package be.iminds.ilabt.jfed.experimenter_gui.editor.properties;

import be.iminds.ilabt.jfed.experimenter_gui.config.AdvertisementFetchManager;
import be.iminds.ilabt.jfed.experimenter_gui.config.JFedGuiConfig;
import be.iminds.ilabt.jfed.experimenter_gui.editor.ModelRspecEditor;
import be.iminds.ilabt.jfed.experimenter_gui.editor.properties.bonfire.BonfireNodePropertiesTab;
import be.iminds.ilabt.jfed.experimenter_gui.editor.properties.map.MapDialogFactory;
import be.iminds.ilabt.jfed.experimenter_gui.health.HealthStatus;
import be.iminds.ilabt.jfed.experimenter_gui.health.MaintenanceMessage;
import be.iminds.ilabt.jfed.experimenter_gui.tasks.TasksFactory;
import be.iminds.ilabt.jfed.experimenter_gui.ui.AvailabilityMeter;
import be.iminds.ilabt.jfed.experimenter_gui.ui.HealthStatusPane;
import be.iminds.ilabt.jfed.experimenter_gui.ui.progressbar.TaskService;
import be.iminds.ilabt.jfed.experimenter_gui.util.BonfireLicenseUtil;
import be.iminds.ilabt.jfed.experimenter_gui.util.StringConverters;
import be.iminds.ilabt.jfed.fedmon.webapi.client.FedmonWebApiClient;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Resource;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.ResourceClass;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Server;

import be.iminds.ilabt.jfed.fedmon.webapi.service.json.ServerGlimpse;
import be.iminds.ilabt.jfed.highlevel.tasks.HighLevelTaskFactory;
import be.iminds.ilabt.jfed.highlevel.util.AggregateManagerWrapperFactory;
import be.iminds.ilabt.jfed.lowlevel.TestbedInfoSource;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityFinder;
import be.iminds.ilabt.jfed.rspec.model.DiskImage;
import be.iminds.ilabt.jfed.rspec.model.SliverType;
import be.iminds.ilabt.jfed.rspec.model.SliverTypeBuilder;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXRspecInterface;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXRspecLink;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXRspecNode;
import be.iminds.ilabt.jfed.ui.javafx.FXMLUtil;
import be.iminds.ilabt.jfed.ui.javafx.GlyphUtils;
import be.iminds.ilabt.jfed.ui.javafx.dialogs.JFDialogs;
import be.iminds.ilabt.jfed.util.GeniUrn;
import com.google.common.collect.Iterables;
import com.google.common.collect.Lists;
import javafx.application.Platform;
import javafx.beans.property.*;
import javafx.beans.value.ChangeListener;
import javafx.beans.value.ObservableValue;
import javafx.collections.FXCollections;
import javafx.collections.ObservableList;
import javafx.event.ActionEvent;
import javafx.fxml.FXML;
import javafx.geometry.Insets;
import javafx.scene.control.*;
import javafx.scene.layout.*;
import javafx.scene.paint.Color;
import javafx.stage.Stage;
import org.controlsfx.glyphfont.FontAwesome;
import org.controlsfx.glyphfont.Glyph;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nullable;
import java.net.MalformedURLException;
import java.net.URL;
import java.time.ZonedDateTime;
import java.util.*;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.stream.Collectors;

/**
 * User: twalcari
 * Date: 4/10/2014
 * Time: 11:55
 */
public class AdaptivePropertiesDialog extends BorderPane {
    private static final Logger LOG = LoggerFactory.getLogger(AdaptivePropertiesDialog.class);

    private static final Comparator<? super MaintenanceMessage> MAINTENANCE_MESSAGE_COMPARATOR =
            Comparator.nullsLast(Comparator.comparing(MaintenanceMessage::getStartTime))
                    .thenComparing(Comparator.nullsLast(Comparator.comparing(MaintenanceMessage::getEndTime)))
                    .thenComparing(MaintenanceMessage::getMessage);

    private final JFedGuiConfig config;
    private final AdvertisementFetchManager advertisementFetchManager;
    private final TestbedInfoSource testbedInfoSource;
    private final AuthorityFinder authorityFinder;

    private final ResourceClass resourceClass;

    private final FXRspecNode rspecNode;
    private final ModelRspecEditor modelRspecEditor;

    private final List<AdaptivePropertiesTab> availableTabs;
    @FXML
    private HealthStatusPane healthStatusPane;
    @FXML
    private VBox propertiesBox;

    @FXML
    private Pane diskImageBox;
    @FXML
    private Button addDiskImageButton;

    @FXML
    private TextField nameTextField;

    @FXML
    private Button saveButton;
    @FXML
    private Button cancelButton;
    @FXML
    private ComboBox<Resource> resourceComboBox;
    @FXML
    private Button testbedsMapButton;

    @FXML
    private Label testbedHealthInfoLabel;

    @FXML
    private TextField diskImageTextField;

    @FXML
    private TabPane tabPane;

    private final Map<Resource, ComponentManagerPropertiesPane> cmiPropertiesPanes = new HashMap<>();
    private final ObjectProperty<ComponentManagerPropertiesPane> currentCmip = new SimpleObjectProperty<>(null);
    private final BooleanProperty isSpecificNodeChosen = new SimpleBooleanProperty();
    private final ReadOnlyObjectProperty<Server> currentServer;
    private final ReadOnlyObjectProperty<Resource> currentResource;
    private final MapDialogFactory mapDialogFactory;
    private final HighLevelTaskFactory hltf;
    private final TasksFactory tasksFactory;
    private final TaskService taskService;
    private final AggregateManagerWrapperFactory aggregateManagerWrapperFactory;

    /* 1 thead that fetches ServerGlimpse in the background whenever needed. */
    private final ExecutorService serverGlimpseFetchExecutor = Executors.newFixedThreadPool(1);
    private final FedmonWebApiClient fedmonWebApiClient;

    private class ResourceCell extends ListCell<Resource> {


        protected final HBox container = new HBox(2);

        protected final Label nameLabel = new Label("");
        protected final Glyph maintenanceIcon =
                GlyphUtils.createGlyph(FontAwesome.Glyph.WRENCH, Color.RED);
        protected final Glyph reputationGlyph = GlyphUtils.createGlyph(FontAwesome.Glyph.STAR, Color.GRAY);
        protected final Glyph healthStatusGlyph = GlyphUtils.createGlyph(FontAwesome.Glyph.HEART, Color.GRAY);

        private final boolean shouldUseVMCalculationMethod;
        @Nullable
        protected final AvailabilityMeter availabilityMeter;

        {
            Pane fillerPane = new Pane();
            HBox.setHgrow(fillerPane, Priority.ALWAYS);

            container.setPadding(new Insets(0, 5, 2, 0));
            container.getChildren().addAll(nameLabel, fillerPane, maintenanceIcon, reputationGlyph, healthStatusGlyph);

            boolean shouldUseVMCalculationMethod = false;
            if (resourceClass != null) {
//                final NodeDescription nodeDescription =
//                        AdaptivePropertiesDialog.this.config.getNodeDescription(node.getResourceClassId());
                if (resourceClass != null) {
                    if (config.getResourceType(resourceClass) == JFedGuiConfig.ResourceType.RAW_PC) {
                        availabilityMeter = new AvailabilityMeter();
                        container.getChildren().add(availabilityMeter);
                    } else if (config.getResourceType(resourceClass) == JFedGuiConfig.ResourceType.VM) {
                        availabilityMeter = new AvailabilityMeter();
                        shouldUseVMCalculationMethod = true;
                        availabilityMeter.setUseVmCalculationMethod(true);
                        container.getChildren().add(availabilityMeter);
                    } else {
                        availabilityMeter = null;
                    }
                } else {
                    availabilityMeter = null;
                }

            } else {
                availabilityMeter = null;
            }

            setGraphic(container);

            this.shouldUseVMCalculationMethod = shouldUseVMCalculationMethod;
        }


        @Override
        protected void updateItem(Resource item, boolean empty) {
            super.updateItem(item, empty);
            if (!empty && item != null) {
                assert item.getServer() != null;
                nameLabel.setText(StringConverters.RESOURCE_STRING_CONVERTER.toString(item));

//                final JFedGuiConfig.TestbedDescription testbedDescription =
//                        AdaptivePropertiesDialog.this.config.getTestbedDescription(item.getAuthority().getDefaultComponentManagerAsGeniUrn());
                HealthStatus healthStatus = config.getHealthStatus(item.getServer());


                if (healthStatus != null) {
                    if (healthStatus.getHealth() != null) {
                        final int health = healthStatus.getHealth();

                        if (health == 100) {
                            healthStatusGlyph.setTextFill(Color.GREEN);
                        } else if (health > 66) {
                            healthStatusGlyph.setTextFill(Color.YELLOWGREEN);
                        } else if (health > 33) {
                            healthStatusGlyph.setTextFill(Color.ORANGE);
                        } else {
                            healthStatusGlyph.setTextFill(Color.RED);
                        }
                        Tooltip.install(healthStatusGlyph, new Tooltip(String.format("%s has a health rating of %d%%",
                                StringConverters.RESOURCE_STRING_CONVERTER.toString(item),
                                health)));

                        healthStatusGlyph.setVisible(true);
                    } else {
                        healthStatusGlyph.setVisible(false);
                    }

                    if (healthStatus.getReputation() != null) {
                        final int reputation = healthStatus.getReputation();

                        if (reputation > 95) {
                            reputationGlyph.setTextFill(Color.GREEN);
                        } else if (reputation > 66) {
                            reputationGlyph.setTextFill(Color.YELLOWGREEN);
                        } else if (reputation > 33) {
                            reputationGlyph.setTextFill(Color.ORANGE);
                        } else {
                            reputationGlyph.setTextFill(Color.RED);
                        }
                        Tooltip.install(reputationGlyph, new Tooltip(String.format("%s has a reputation rating of %d%%",
                                StringConverters.RESOURCE_STRING_CONVERTER.toString(item),
                                reputation)));


                        reputationGlyph.setVisible(true);
                    } else {
                        reputationGlyph.setVisible(false);
                    }

                    if (healthStatus.getMaintenance() != null) {
                        List<MaintenanceMessage> activeMaintenances =
                                healthStatus.getMaintenance().stream()
                                        .filter(mm -> mm.getStartTime() == null || mm.getStartTime().isBefore(ZonedDateTime.now()))
                                        .filter(mm -> mm.getEndTime() == null || mm.getEndTime().isAfter(ZonedDateTime.now()))
                                        .sorted(MAINTENANCE_MESSAGE_COMPARATOR)
                                        .collect(Collectors.toList());

                        List<MaintenanceMessage> futureMaintenances =
                                healthStatus.getMaintenance().stream()
                                        .filter(mm -> mm.getStartTime() != null && mm.getStartTime().isAfter(ZonedDateTime.now()))
                                        .sorted(MAINTENANCE_MESSAGE_COMPARATOR)
                                        .collect(Collectors.toList());

                        if (!activeMaintenances.isEmpty()) {
                            Tooltip.install(maintenanceIcon, new Tooltip(MaintenanceMessage.formatMaintenanceMessages(activeMaintenances)));
                            maintenanceIcon.setTextFill(Color.RED);
                            maintenanceIcon.setVisible(true);
                        } else if(!futureMaintenances.isEmpty()) {
                            Tooltip.install(maintenanceIcon, new Tooltip(MaintenanceMessage.formatMaintenanceMessages(futureMaintenances)));
                            maintenanceIcon.setTextFill(Color.ORANGE);
                            maintenanceIcon.setVisible(true);
                        }else{
                            maintenanceIcon.setVisible(false);
                        }
                    } else {
                        maintenanceIcon.setVisible(false);
                    }

                    if (availabilityMeter != null) {
                        if (shouldUseVMCalculationMethod
                                && healthStatus.getVMsAvailable() != null
                                && healthStatus.getVMsTotal() != null
                                && healthStatus.getVMsTotal() != 0) {
                            availabilityMeter.setUseVmCalculationMethod(true);

                            availabilityMeter.setAvailable(healthStatus.getVMsAvailable());
                            availabilityMeter.setTotal(healthStatus.getVMsTotal());
                            availabilityMeter.setVisible(true);
                            availabilityMeter.setUndetermined(false);

                            Tooltip.install(availabilityMeter,
                                    new Tooltip(healthStatus.getVMsAvailable() + " of "
                                            + healthStatus.getVMsTotal() + " VM available"));
                        } else {
                            availabilityMeter.setUseVmCalculationMethod(false);
                            if (healthStatus.getRawPCsAvailable() != null
                                    && healthStatus.getRawPCsTotal() != null
                                    && healthStatus.getRawPCsTotal() != 0) {
                                availabilityMeter.setAvailable(healthStatus.getRawPCsAvailable());
                                availabilityMeter.setTotal(healthStatus.getRawPCsTotal());
                                availabilityMeter.setVisible(true);
                                availabilityMeter.setUndetermined(false);

                                Tooltip.install(availabilityMeter,
                                        new Tooltip(healthStatus.getRawPCsAvailable() + " of "
                                                + healthStatus.getRawPCsTotal() + " bare metal available"));

                                availabilityMeter.setVisible(true);
                            }else if(healthStatus.getAnyAvailable() != null
                                    && healthStatus.getAnyTotal() != null
                                    && healthStatus.getAnyTotal() != 0) {
                                availabilityMeter.setAvailable(healthStatus.getAnyAvailable());
                                availabilityMeter.setTotal(healthStatus.getAnyTotal());
                                availabilityMeter.setVisible(true);
                                availabilityMeter.setUndetermined(false);

                                Tooltip.install(availabilityMeter,
                                        new Tooltip(healthStatus.getAnyAvailable() + " of "
                                                + healthStatus.getAnyTotal() + " resources available"));
                            }    else{
                                availabilityMeter.setAvailable(0);
                                availabilityMeter.setTotal(0);
                                availabilityMeter.setVisible(true);
                                availabilityMeter.setUndetermined(true);
                            }
                        }
                    }


                } else {
                    healthStatusGlyph.setVisible(false);
                    reputationGlyph.setVisible(false);
                    maintenanceIcon.setVisible(false);
                    if (availabilityMeter != null)
                        availabilityMeter.setVisible(false);
                }


            } else {
                nameLabel.setText("");
                healthStatusGlyph.setVisible(false);
                reputationGlyph.setVisible(false);
                maintenanceIcon.setVisible(false);
                if (availabilityMeter != null)
                    availabilityMeter.setVisible(false);
            }
        }

    }

    AdaptivePropertiesDialog(ModelRspecEditor modelRspecEditor, final FXRspecNode rspecNode, JFedGuiConfig config, AdvertisementFetchManager advertisementFetchManager,
                             TestbedInfoSource testbedInfoSource, AuthorityFinder authorityFinder, ResourceClass resourceClass, MapDialogFactory mapDialogFactory,
                             HighLevelTaskFactory hltf, TasksFactory tasksFactory, TaskService taskService,
                             AggregateManagerWrapperFactory aggregateManagerWrapperFactory, BonfireLicenseUtil bonfireLicenseUtil,
                             FedmonWebApiClient fedmonWebApiClient) {
        this.config = config;
        this.advertisementFetchManager = advertisementFetchManager;
        this.testbedInfoSource = testbedInfoSource;
        this.authorityFinder = authorityFinder;
        this.resourceClass = resourceClass;
        this.mapDialogFactory = mapDialogFactory;
        this.hltf = hltf;
        this.tasksFactory = tasksFactory;
        this.taskService = taskService;
        this.aggregateManagerWrapperFactory = aggregateManagerWrapperFactory;
        this.fedmonWebApiClient = fedmonWebApiClient;

        assert resourceClass != null;
        assert modelRspecEditor != null;
        assert rspecNode != null;

        this.modelRspecEditor = modelRspecEditor;
        this.rspecNode = rspecNode;


        //initialize the Scene
        FXMLUtil.injectFXML(this);

        assert nameTextField != null;
        assert resourceComboBox != null;

        //initialize fields of main tab
        nameTextField.setTooltip(
                new Tooltip("Only alphanumerical, max. " + ModelRspecEditor.MAX_NODE_NAME_LENGTH + " characters"));


        //set initial values to these of the received node
        nameTextField.setText(rspecNode.getClientId());

        healthStatusPane.setType(config.getResourceType(resourceClass));

        if (rspecNode.getResourceClassId() != null) {
            ResourceClass nodeResourceClass = config.getResourceClass(rspecNode.getResourceClassId());
            if (nodeResourceClass != null) {
                assert nodeResourceClass.getId().equals(resourceClass.getId());
            }
        }
        rspecNode.setResourceClassId(resourceClass.getId());

        //populate tabPane
        availableTabs = new ArrayList<>();

        BootScriptsTab bootScriptsTab = new BootScriptsTab(this);
        bootScriptsTab.load(rspecNode);
        availableTabs.add(bootScriptsTab);

        ExoGeniConfigTab exoGeniConfigTab = new ExoGeniConfigTab(this);
        exoGeniConfigTab.load(rspecNode);
        availableTabs.add(exoGeniConfigTab);

        RoutableIPTab routableIPTab = new RoutableIPTab(this);
        routableIPTab.load(rspecNode);
        availableTabs.add(routableIPTab);

        XenTab xenTab = new XenTab(this);
        xenTab.load(rspecNode);
        availableTabs.add(xenTab);

        //add tab for each interface. Can only be done after nameTextField has been properly initialized!
        for (FXRspecInterface iface : rspecNode.getInterfaces()) {
            tabPane.getTabs().add(new InterfacePropertiesTab(modelRspecEditor, rspecNode, iface, nameTextField.textProperty()));
        }

        assert resourceClass.getResources() != null;
        ObservableList<Resource> resources = FXCollections.observableArrayList(resourceClass.getResources());

        assert resourceComboBox != null;
        resourceComboBox.setItems(resources);
        resourceComboBox.setConverter(StringConverters.RESOURCE_STRING_CONVERTER);
        resourceComboBox.setCellFactory(param -> new ResourceCell());
        resourceComboBox.setButtonCell(new ResourceCell() {

            {
                nameLabel.setTextFill(Color.BLACK);
                container.setPadding(new Insets(0, 0, 0, 0));
            }

        });

        this.currentCmip.addListener(new ChangeListener<ComponentManagerPropertiesPane>() {
            @Override
            public void changed(ObservableValue<? extends ComponentManagerPropertiesPane> observable, ComponentManagerPropertiesPane oldValue, ComponentManagerPropertiesPane newValue) {
                isSpecificNodeChosen.unbind();
                if (newValue != null) {
                    isSpecificNodeChosen.bind(newValue.isSpecificNodeChosen());
                } else
                    isSpecificNodeChosen.setValue(false);
            }
        });


        this.currentResource = resourceComboBox.getSelectionModel().selectedItemProperty();
        SimpleObjectProperty currentServerSimpleObjectProperty = new SimpleObjectProperty<>();
        this.currentServer = currentServerSimpleObjectProperty;
        currentResource.addListener(new ChangeListener<Resource>() {
            @Override
            public void changed(ObservableValue<? extends Resource> observable, Resource oldValue, Resource newValue) {
                if (newValue == null)
                    currentServerSimpleObjectProperty.set(null);
                else
                    currentServerSimpleObjectProperty.set(newValue.getServer());
            }
        });

        this.currentResource.addListener(new ChangeListener<Resource>() {
            @Override
            public void changed(ObservableValue<? extends Resource> observableValue,
                                Resource oldValue, Resource newValue) {

                boolean activeNodesOnly = true;

                if (oldValue != null) {
                    Resource resource = oldValue;

                    if (resource.getAllowDiskImage() &&
                            resource.getDiskImage() != null &&
                            resource.getDiskImage().equals(diskImageTextField.getText())) {
                        diskImageTextField.setText("");
                    }

                    activeNodesOnly = cmiPropertiesPanes.get(oldValue).isActiveNodesOnly();
                }

                if (newValue != null) {
                    Resource resource = newValue;
                    assert resource.getServer() != null;

                    if (!cmiPropertiesPanes.containsKey(newValue)) {
                        Runnable fetchServerGlimpse = () -> {
                            try {
                                LOG.debug("Fetching ServerGlimpse in background");
                                try {
                                    //Note: FedmonWebApiCachedClient caches ServerGlimpse for 10 minutes with default settings
                                    assert resource != null;
                                    assert resource.getServer() != null;
                                    assert resource.getServer().getId() != null;
                                    final Optional<ServerGlimpse> serverGlimpse =
                                            fedmonWebApiClient.getServerGlimpseByServerId(resource.getServer().getId());
                                    if (serverGlimpse.isPresent()) {
                                        LOG.debug("Got ServerGlimpse in background");
                                        if (serverGlimpse.get() != null) {
                                            Platform.runLater(() -> {
                                                config.setServerGlimpse(resource.getServer().getId(), serverGlimpse.get());
                                                //TODO register that fetching ServerGlimpse is done (so advertisement fetch can start if needed)
                                            });
                                        } else {
                                            assert serverGlimpse.get() != null;
                                            LOG.error("ServerGlimpse never be null for server " + resource.getServer().getId());
                                                //TODO register that fetching ServerGlimpse is done (so advertisement fetch can start if needed)
                                        }
                                    } else {
                                        LOG.debug("Did NOT get ServerGlimpse in background");
                                        //TODO register that fetching ServerGlimpse is done (so advertisement fetch can start if needed)
                                    }
                                } catch (FedmonWebApiClient.FedmonWebApiClientException e) {
                                    LOG.error("Error fetching ServerGlimpse for server " + resource.getServer().getId(), e);
                                }
                            } catch (Throwable t) {
                                LOG.error("Unexpected Exception fetching ServerGlimpse", t);
                                throw t;
                            }
                        };
                        serverGlimpseFetchExecutor.submit(fetchServerGlimpse);

                        cmiPropertiesPanes.put(currentResource.get(), new ComponentManagerPropertiesPane(resource, AdaptivePropertiesDialog.this.rspecNode,
                                activeNodesOnly, AdaptivePropertiesDialog.this.config, AdaptivePropertiesDialog.this.advertisementFetchManager, mapDialogFactory, bonfireLicenseUtil));
                    }
                    propertiesBox.getChildren().add(cmiPropertiesPanes.get(newValue));
                    currentCmip.set(cmiPropertiesPanes.get(newValue));

                    if (resource.getAllowDiskImage() &&
                            resource.getDiskImage() != null &&
                            diskImageTextField.getText().isEmpty()) {
                        diskImageTextField.setText(resource.getDiskImage());
                    }
                    if (!resource.getAllowDiskImage()) {
//                        diskImageTextField.setText("");
                        diskImageBox.setVisible(false);
                    } else {
                        diskImageBox.setVisible(true);
                    }

                    HealthStatus healthStatus = AdaptivePropertiesDialog.this.config.getHealthStatus(resource.getServer());
                    healthStatusPane.setHealthStatus(healthStatus);
                }
                else
                    currentCmip.set(null);

                if (oldValue != null) {
                    propertiesBox.getChildren().remove(cmiPropertiesPanes.get(oldValue));
                }

                if (newValue != null) {
                    //only show disk image button for instageni and emulab-testbeds
                    Server server = newValue.getServer();
                    boolean showDiskImageDialog =
//                            server.isServerType("instageni") || server.isServerType("emulab") ||
                            server.hasFlag(Server.Flag.featureListDiskImages);
                    addDiskImageButton.setVisible(showDiskImageDialog);
                }

                determineTabsVisibility(rspecNode);
            }
        });

        Resource resource = config.findBestResource(resourceClass, rspecNode);

        if (resource == null) {
            //no component manager: select nothing
            resourceComboBox.getSelectionModel().clearSelection();
        } else {
            //select the current testbed in the list
            resourceComboBox.getSelectionModel().select(resource);
        }

        if (!rspecNode.getSliverTypes().isEmpty()) {
            List<DiskImage> diskImages = rspecNode.getSliverTypes().get(0).getDiskImages();

            if (!diskImages.isEmpty())
                diskImageTextField.setText(diskImages.get(0).getName() != null ?
                        diskImages.get(0).getName() :
                        diskImages.get(0).getUrl());
        }

        diskImageBox.managedProperty().bind(diskImageBox.visibleProperty());
        addDiskImageButton.managedProperty().bind(addDiskImageButton.visibleProperty());

        //only show maps-button if running on a 64-bit JVM (see FEDIBBTDEV-1344)
        String arch = System.getProperty("os.arch");

        if (arch.equals("x86")) {
            testbedsMapButton.setVisible(false);
        }

        //do this when currentCmi is initialised and set to current CMI
        BonfireNodePropertiesTab bonfireNodePropertiesTab = new BonfireNodePropertiesTab(this);
        bonfireNodePropertiesTab.load(rspecNode);
        availableTabs.add(bonfireNodePropertiesTab);

        determineTabsVisibility(rspecNode);
    }

    @FXML
    private void showTestbedsMap() {
        Resource selectedResource =
                mapDialogFactory.showTestbedMapDialog(getScene().getWindow(), resourceClass);
        if (selectedResource != null)
            resourceComboBox.getSelectionModel().select(selectedResource);
    }

    /**
     * This function shows and hides tabs as needed. It is called by evenListeners on the testbeds- and node-comboboxes.
     */
    private void determineTabsVisibility(FXRspecNode rspecNode) {
        int insertPosition = 1;
        for (AdaptivePropertiesTab apt : availableTabs) {
            if (apt.shouldBeVisible(rspecNode)) {
                LOG.debug("Tab {} should be made visible.", apt.getClass().getSimpleName());
                if (tabPane.getTabs().contains(apt)) {
                    insertPosition++;
                } else {
                    tabPane.getTabs().add(insertPosition, apt);
                }
            } else {
                LOG.debug("Tab {} should be made invisible.", apt.getClass().getSimpleName());
                if (tabPane.getTabs().contains(apt)) {
                    tabPane.getTabs().remove(apt);
                }
            }
        }
    }

    @FXML
    @SuppressWarnings("unchecked")
    public void onSaveAction(ActionEvent actionEvent) {
        nameTextField.setText(nameTextField.getText().trim());

        if (!ModelRspecEditor.isValidNodeName(nameTextField.getText())) {
            JFDialogs.create()
                    .owner(this)
                    .message("A node name may only contain alphanumerical characters, and is maximum " + ModelRspecEditor.MAX_NODE_NAME_LENGTH + " characters long.")
                    .masthead("The entered node name is invalid.")
                    .showError();
            return;
        }

        if (currentResource.get() == null) {
            JFDialogs.create()
                    .owner(this)
                    .message("You must select an authority for this node.")
                    .masthead("No authority selected")
                    .title("No authority selected")
                    .showError();
            return;
        }


        if (!diskImageTextField.getText().isEmpty()) {
            //first remove any extra spaces that have slipped in
            diskImageTextField.setText(diskImageTextField.getText().trim().replace(" ", ""));

            String diskImageValue = diskImageTextField.getText();

            //now check if it is a valid URN or HTTP(S)-link

            if (diskImageValue.startsWith("urn")) {
                GeniUrn diskImageUrn = GeniUrn.parse(diskImageValue);

                if (diskImageUrn == null || !"image".equals(diskImageUrn.getEncodedResourceType())) {
                    JFDialogs.create()
                            .owner(this)
                            .message("The provided disk image URN is invalid. Please correct the entered value.")
                            .masthead("Invalid disk image URN")
                            .showError();

                    LOG.warn("User entered invalid disk image URN '{}'", diskImageValue);

                    return;
                }

            } else {
                //we expect a valid URL instead
                try {
                    new URL(diskImageValue);
                } catch (MalformedURLException e) {
                    JFDialogs.create()
                            .owner(this)
                            .message("The provided disk image URL is invalid. Please correct the entered value.")
                            .masthead("Invalid disk image URL")
                            .showError();


                    LOG.warn("User entered invalid disk image URL '{}'", diskImageValue);

                    return;
                }
            }
        }

        ComponentManagerPropertiesPane componentManagerPropertiesPane = cmiPropertiesPanes.get(currentResource.get());

        if (!componentManagerPropertiesPane.validate())
            return;


        for (AdaptivePropertiesTab visibleTab : Iterables.filter(tabPane.getTabs(), AdaptivePropertiesTab.class)) {
            String validation = visibleTab.validateConfiguration();

            if (validation != null) {
                JFDialogs.create()
                        .owner(this)
                        .message(validation)
                        .masthead("Node configuration could not be saved")
                        .showError();
                return;
            }

        }


        for (PropertiesTab visibleTab : Iterables.filter(tabPane.getTabs(), PropertiesTab.class)) {
            String validation = visibleTab.validateConfiguration();

            if (validation != null) {
                JFDialogs.create()
                        .owner(this)
                        .message(validation)
                        .masthead("Node configuration could not be saved")
                        .showError();
                return;
            }

        }


        FXRspecNode currentNode = this.rspecNode;

        if (!nameTextField.getText().equals(currentNode.getClientId())) {
            modelRspecEditor.updateNodeClientId(currentNode, nameTextField.getText());
        }

        Resource resource = resourceComboBox.getSelectionModel().getSelectedItem();
        Server server = resource == null ? null : resource.getServer();

        boolean componentManagerChanged = !server.getDefaultComponentManagerAsGeniUrn().equals(currentNode.getComponentManagerId());
        currentNode.setComponentManagerId(server.getDefaultComponentManagerAsGeniUrn());

        //force SliverType to an existing type on the testbed
        //NOTE: this is a good default, that gets overwritten when the ExoGeni tab is present and is saved later in this step
        if (resource.getSliverType() != null &&
                (currentNode.getSliverTypeName() == null || !resource.getSliverType().equals(currentNode.getSliverTypeName()))
                ) {
            SliverType resourceSliverType = new SliverTypeBuilder(resource.getSliverType()).build();
//            if (resourceSliverType != null) {
            //TODO post fedmon-v3 refactor: I don't think this has any use: we could check the advertisement info, but it is ok to just trust the Resource info
//                if (!server.getSliverTypes().isEmpty() && !server.hasSliverType(resourceSliverType.getName())) {
//                    LOG.warn("Default sliver_type for " + server.getDefaultComponentManagerAsGeniUrn() + " is \"" + resourceSliverType.getName() + "\", but that sliver_type is not in ListResources. Picking an existing one instead.");
//                    resourceSliverType = null;
//                } else {
//                    LOG.warn("Default sliver_type for " + server.getDefaultComponentManagerAsGeniUrn() + " is \"" + resourceSliverType.getName() + "\", but that sliver_type is not in ListResources. However, no alternative is found in ListResources...");
//                }
//            }

            //TODO post fedmon-v3 refactor: see above, this might not be needed.
//            if (resourceSliverType == null && !server.getSliverTypes().isEmpty())
//                resourceSliverType = server.getSliverTypes().get(0); //choose first (= quite randomly)

            if (resourceSliverType != null)
                currentNode.setSliverTypeName(resourceSliverType.getName());
        }

        currentNode.setComponentId(
                componentManagerPropertiesPane.getSelectedComponentId() != null
                        ? componentManagerPropertiesPane.getSelectedComponentId()
                        : null);

        if (componentManagerPropertiesPane.getSelectedHardwareType() != null)
            currentNode.getHardwareTypes().setAll(componentManagerPropertiesPane.getSelectedHardwareType());


        if (componentManagerChanged) {
            final boolean supportsStitching =
                    server.hasFlag(Server.Flag.featureStitching);
            List<FXRspecLink> toRemove = new ArrayList<>();

            if (!server.getTestbed().getAllowLinks() && rspecNode.getInterfaces().size() > 0) {
                JFDialogs.create()
                        .owner(this)
                        .message(String.format("'%s' does not support dedicated links. Existing links to '%s' are being removed.",
                                server.getName(),
                                rspecNode.getClientId()))
                        .masthead("Illegal links detected")
                        .title("Illegal links detected")
                        .showWarning();

                toRemove.addAll(Lists.transform(rspecNode.getInterfaces(),
                        rspecInterface -> rspecInterface.isLinkBound() ? rspecInterface.getLink() : null
                ));
            } else {
                for (FXRspecInterface iface : currentNode.getInterfaces()) {
                    if (!iface.isLinkBound())
                        continue;

                    if (iface.getLink().isStitched(testbedInfoSource)) {

                        if (!isValidStitchedLink(iface.getLink())) {

                            Alert alert =
                                    new Alert(Alert.AlertType.WARNING,
                                            String.format("By changing the authority of this node, the link '%s' has become a stitched link.\n" +
                                                            "However, one or more authorities do not support stitched links. It is therefore removed.",
                                                    iface.getLink().getClientId()),
                                            ButtonType.OK);
                            alert.setHeaderText("Invalid stitched link");
                            alert.setTitle("Invalid stitched link");
                            alert.show();

                            toRemove.add(iface.getLink());
                        } else if (JFDialogs.create()
                                .owner(this)
                                .message(
                                        "Link '" + iface.getLink().getClientId() + "' is now a stitched link.\n\n" +
                                                "Stitched links between testbeds are currently VERY experimental and not supported." +
                                                " Do you want to remove this link?")
                                .masthead("Experimental functionality detected.")
                                .title("Experimental functionality detected.")
                                .buttonTypes(ButtonType.YES, ButtonType.NO)
                                .showWarning().filter(r -> r == ButtonType.YES).isPresent()) {


                            toRemove.add(iface.getLink());

                        } else {

                            modelRspecEditor.convertLinkToStitchedLink(iface.getLink());
                        }
                    } else {
                        //the link used to be stitched, but isn't anymore
                        JFDialogs.create()
                                .owner(this)
                                .message("Link '" + iface.getLink().getClientId() + "' is now a normal link." +
                                        " Some settings are reverted to default.")
                                .masthead("Link type change detected")
                                .title("Link type change detected")
                                .showInformation();

                        modelRspecEditor.convertLinkToNormalLink(iface.getLink());

                    }
                }
            }

            toRemove.forEach(modelRspecEditor.getModelRspec()::deleteLink);
        }

        if (!currentNode.getSliverTypes().isEmpty()) {
            SliverType sliverType = currentNode.getSliverTypes().get(0);

            if (sliverType != null) {
                assert diskImageTextField != null;
                if (diskImageTextField.getText() != null) {
                    String diskImageString = diskImageTextField.getText().trim();

                    if (!diskImageString.isEmpty()) {
                        currentNode.setSliverTypeDiskImage(diskImageString);
                    } else {
                        currentNode.setSliverTypeDiskImage(null);
                    }
                } else {
                    currentNode.setSliverTypeDiskImage(null);
                }
            }
        } else {
            LOG.error("Could not find sliver-type on currentNode: cannot save disk image");
        }

        //check if we need to change the node exclusivity
        Boolean exclusive =
                config.getNodeExclusivity(currentNode.getComponentManagerId(), currentNode.getSliverTypeName());

        if (exclusive != null) {
            currentNode.setExclusive(exclusive);
        }

        final FXRspecNode resultNode = currentNode;

        tabPane.getTabs().stream()
                .filter(tab -> tab instanceof AdaptivePropertiesTab)
                .map(tab -> (AdaptivePropertiesTab) tab).forEach(tab -> {
            assert (tab.shouldBeVisible(resultNode)) :
                    tab.getClass().getName() + "indicates it shouldn't be visible at this time!";

            //we assume here that when the tab is visible, it is also able to save it's information
            // to the new FXRspecNode-class!
            // THIS IS POTENTIALLY UNSAFE!
            tab.save(resultNode);
        });


        tabPane.getTabs().stream()
                .filter(tab -> tab instanceof PropertiesTab)
                .map(tab -> (PropertiesTab) tab)
                .forEach(PropertiesTab::save);


        ResourceClass resourceClass = config.findBestResourceClass(currentNode);
        if (resourceClass != null)
            currentNode.setResourceClassId(resourceClass.getId());


        Stage stage = (Stage) saveButton.getScene().getWindow();
        stage.close();
    }

    String getCurrentSliverTypeName() {
        //get the Resource for the authority that is currently visible
        Resource resource = resourceComboBox.getSelectionModel().getSelectedItem();

        if (resource == null)
            return null;

        if (rspecNode.getSliverTypeName() != null &&
                resource.getSliverType() != null && resource.getSliverType().equals(rspecNode.getSliverTypeName())
                ) {
            return rspecNode.getSliverTypeName();
        }

        //TODO: check if rspecNode.getSliverTypeName() is valid for this server and change resourceType if needed?

        //fallback onto the default sliver-type defined in the CMI
        String fallbackSliverType = resource.getSliverType();

        if (fallbackSliverType != null) {
            return fallbackSliverType;
        }

        //TODO: check if this functionality should be restored
//        LOG.warn("Default sliver_type for {} is '{}', but that sliver_type is not in ListResources. " +
//                        "Picking an existing one instead.",
//                server.getName(), fallbackSliverType);

//        if (!server.getSliverTypes().isEmpty())
//            return server.getSliverTypes().get(0).getName(); //choose first (= quite randomly)
//        else
        return null;
    }


    private boolean isValidStitchedLink(FXRspecLink link) {
        for (FXRspecInterface iface : link.getInterfaces()) {
            if (!iface.isNodeBound())
                continue;


            Server nodeAuth =
                    authorityFinder.findByUrn(iface.getNode().getComponentManagerId(), AuthorityFinder.Purpose.STITCHING);

            if (!nodeAuth.hasFlag(Server.Flag.featureStitching))
                return false;
        }

        return true;
    }


    @FXML
    public void onCancelAction(ActionEvent actionEvent) {
        Stage stage = (Stage) cancelButton.getScene().getWindow();
        stage.close();
    }

    @FXML
    private void onRequestDiskImageAction() {

        String newDiskImage =
                new SelectDiskImageDialog(this.getScene().getWindow(), resourceComboBox.getValue(),
                        config, testbedInfoSource, advertisementFetchManager,
                        hltf, tasksFactory, taskService, aggregateManagerWrapperFactory).showDialog();

        if (newDiskImage != null) {
            diskImageTextField.setText(newDiskImage);
        }
    }

    /**
     * @return may return null if nothing is selected
     */
    public Resource getSelectedResource() {
        return currentResource.get();
    }
    public Server getSelectedServer() {
        return currentServer.get();
    }

    public GeniUrn getSelectedComponentId() {
        return cmiPropertiesPanes.containsKey(currentServer.get()) ?
                cmiPropertiesPanes.get(currentServer.get()).getSelectedComponentId() : null;
    }

    public ReadOnlyObjectProperty<Resource> selectedResourceProperty() {
        return currentResource;
    }

    public ReadOnlyBooleanProperty isSpecificNodeChosenProperty() {
        return isSpecificNodeChosen;
    }

    public JFedGuiConfig getConfig() {
        return config;
    }

    public AdvertisementFetchManager getAdvertisementFetchManager() {
        return advertisementFetchManager;
    }

    public ResourceClass getResourceClass() {
        return resourceClass;
    }
    public String getResourceClassId() {
        return resourceClass == null ? null : resourceClass.getId();
    }
}
