package be.iminds.ilabt.jfed.experimenter_gui.editor.properties;

import be.iminds.ilabt.jfed.experimenter_gui.config.AdvertisementFetchManager;
import be.iminds.ilabt.jfed.experimenter_gui.config.JFedGuiConfig;
import be.iminds.ilabt.jfed.experimenter_gui.editor.ModelRspecEditor;
import be.iminds.ilabt.jfed.experimenter_gui.editor.properties.map.MapDialogFactory;
import be.iminds.ilabt.jfed.experimenter_gui.tasks.TasksFactory;
import be.iminds.ilabt.jfed.experimenter_gui.ui.progressbar.TaskService;
import be.iminds.ilabt.jfed.experimenter_gui.util.BonfireLicenseUtil;
import be.iminds.ilabt.jfed.experimenter_gui.util.ui.FXUtils;
import be.iminds.ilabt.jfed.experimenter_gui.util.ui.StageUtils;

import be.iminds.ilabt.jfed.fedmon.webapi.client.FedmonWebApiClient;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.ResourceClass;
import be.iminds.ilabt.jfed.highlevel.tasks.HighLevelTaskFactory;
import be.iminds.ilabt.jfed.highlevel.util.AggregateManagerWrapperFactory;
import be.iminds.ilabt.jfed.lowlevel.TestbedInfoSource;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityFinder;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXRspecNode;
import javafx.scene.Scene;
import javafx.stage.Stage;
import javafx.stage.Window;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.inject.Inject;
import javax.inject.Singleton;
import java.util.List;

/**
 * User: twalcari
 * Date: 12/2/2014
 * Time: 14:45
 */
@Singleton
public class AdaptivePropertiesDialogFactory {
    private static final Logger LOG = LoggerFactory.getLogger(AdaptivePropertiesDialogFactory.class);

    private final JFedGuiConfig config;
    private final TestbedInfoSource testbedInfoSource;
    private final AdvertisementFetchManager advertisementFetchManager;
    private final AuthorityFinder authorityFinder;
    private final MapDialogFactory mapDialogFactory;
    private final HighLevelTaskFactory hltf;
    private final TasksFactory tasksFactory;
    private final TaskService taskService;
    private final AggregateManagerWrapperFactory aggregateManagerWrapperFactory;
    private final BonfireLicenseUtil bonfireLicenseUtil;
    private final FedmonWebApiClient fedmonWebApiClient;

    //Note: FedmonWebApiClient is bound to singleton FedmonWebApiCachedClient in HighLevelModule
    //Note: FedmonWebApiCachedClient caches ServerGlimpse for 10 minutes with default settings

    @Inject
    public AdaptivePropertiesDialogFactory(JFedGuiConfig config, TestbedInfoSource testbedInfoSource,
                                           AdvertisementFetchManager advertisementFetchManager, AuthorityFinder authorityFinder,
                                           MapDialogFactory mapDialogFactory,
                                           HighLevelTaskFactory hltf, TasksFactory tasksFactory, TaskService taskService,
                                           AggregateManagerWrapperFactory aggregateManagerWrapperFactory, BonfireLicenseUtil bonfireLicenseUtil,
                                           FedmonWebApiClient fedmonWebApiClient) {
        this.config = config;
        this.testbedInfoSource = testbedInfoSource;
        this.advertisementFetchManager = advertisementFetchManager;
        this.authorityFinder = authorityFinder;
        this.mapDialogFactory = mapDialogFactory;
        this.hltf = hltf;
        this.tasksFactory = tasksFactory;
        this.taskService = taskService;
        this.aggregateManagerWrapperFactory = aggregateManagerWrapperFactory;
        this.bonfireLicenseUtil = bonfireLicenseUtil;
        this.fedmonWebApiClient = fedmonWebApiClient;
    }

    public AdaptivePropertiesDialog createAdaptivePropertiesDialog(ModelRspecEditor modelRspecEditor, FXRspecNode node) {
        ResourceClass resourceClass =
                node.getResourceClassId() == null ?
                        null :
                        config.getResourceClass(node.getResourceClassId());

        List<ResourceClass> matchingResourceClasses = config.findAllPossibleResourceClasses(node);

        if (resourceClass != null && !matchingResourceClasses.contains(resourceClass)) {
            resourceClass = config.findBestResourceClass(node);
            LOG.debug("ResourceClass know for node "+node.getClientId()+" is \""+resourceClass.getId()+"\". " +
                    "But that class does not match the node -> changing it to: "+resourceClass.getId());
        }

        if (resourceClass == null) {
            resourceClass = config.findBestResourceClass(node);
            LOG.debug("No resourceClass know for node "+node.getClientId()+". Falling back to: "+resourceClass.getId());
        }

        assert resourceClass != null;

        return new AdaptivePropertiesDialog(modelRspecEditor, node,
                config, advertisementFetchManager, testbedInfoSource,
                authorityFinder, resourceClass, mapDialogFactory,
                hltf, tasksFactory, taskService, aggregateManagerWrapperFactory, bonfireLicenseUtil, fedmonWebApiClient);
    }

    public void showAdaptivePropertiesDialog(Window parent,
                                             ModelRspecEditor modelRspecEditor, FXRspecNode node) {
        AdaptivePropertiesDialog adaptivePropertiesDialog = createAdaptivePropertiesDialog(modelRspecEditor, node);
        Scene scene = new Scene(adaptivePropertiesDialog);

        Stage dialogStage = new Stage();
        StageUtils.setJFedStageIcons(dialogStage);
        dialogStage.setScene(scene);

        String debug = "";
        int addDebug = 0;
        assert (addDebug++)==0;
        if (addDebug > 0) debug = " ("+adaptivePropertiesDialog.getResourceClassId()+")";

        dialogStage.setTitle("Properties of " + node.getClientId()+debug);
        dialogStage.setMinHeight(500);
        dialogStage.setMaxHeight(600);
        dialogStage.setMinWidth(500);
        FXUtils.centerToOwner(parent, dialogStage);

        dialogStage.showAndWait();
    }
}
