package be.iminds.ilabt.jfed.experimenter_gui.editor.properties;

import be.iminds.ilabt.jfed.experimenter_gui.config.AdvertisementFetchManager;
import be.iminds.ilabt.jfed.experimenter_gui.config.JFedGuiConfig;
import be.iminds.ilabt.jfed.experimenter_gui.config.JFedGuiConfigImpl;
import be.iminds.ilabt.jfed.experimenter_gui.editor.properties.map.ImageMap;
import be.iminds.ilabt.jfed.experimenter_gui.editor.properties.map.MapDialogFactory;
import be.iminds.ilabt.jfed.experimenter_gui.util.BonfireLicenseUtil;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Resource;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Server;
import be.iminds.ilabt.jfed.lowlevel.JFedConnectivityException;
import be.iminds.ilabt.jfed.rspec.model.HardwareType;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXRspecNode;
import be.iminds.ilabt.jfed.ui.javafx.FXMLUtil;
import be.iminds.ilabt.jfed.ui.javafx.dialogs.JFDialogs;
import be.iminds.ilabt.jfed.util.GeniUrn;
import javafx.beans.property.ObjectProperty;
import javafx.beans.property.ReadOnlyBooleanProperty;
import javafx.beans.property.SimpleObjectProperty;
import javafx.beans.value.ChangeListener;
import javafx.beans.value.ObservableValue;
import javafx.event.ActionEvent;
import javafx.fxml.FXML;
import javafx.scene.control.*;
import javafx.scene.layout.StackPane;
import javafx.scene.layout.VBox;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Comparator;
import java.util.Optional;

/**
 * User: twalcari
 * Date: 8/29/2014
 * Time: 11:26
 */
public class ComponentManagerPropertiesPane extends StackPane {
    private final static Logger LOG = LoggerFactory.getLogger(ComponentManagerPropertiesPane.class);

    //moved to AdvertisementFetchManager.isHardwareTypeToIgnore
//    /**
//     * Blacklists as mentioned in FEDIBBTDEV-2100
//     */
//    private static final ImmutableSet<String> HARDWARE_TYPE_NAMES_SUBSTRING_BLACKLIST =
//            ImmutableSet.of("80211", "flex900", "xcvr2450", "delay", "interconnect", "ipv4", "lan", "vm", "switch");
//    private static final ImmutableSet<String> HARDWARE_TYPE_NAMES_BLACKLIST =
//            ImmutableSet.of("pc");

    private static final Comparator<HardwareType> HARDWARE_TYPE_COMPARATOR =
            Comparator.comparing(HardwareType::getName, String::compareToIgnoreCase);

    private final JFedGuiConfig config;
    private final AdvertisementFetchManager advertisementFetchManager;
    private final MapDialogFactory mapDialogFactory;
    private final Resource resource;
    //    private final JFedGuiConfig.TestbedDescription testbedDescription;
    private final BonfireLicenseUtil bonfireLicenseUtil;

    @FXML
    private ComboBox<JFedGuiConfig.ComponentInfo> componentIdComboBox;
    @FXML
    private ComboBox<JFedGuiConfig.HardwareTypeInfo> hardwareTypesComboBox;
    @FXML
    private ToggleGroup selectComponentIdToggleGroup;
    @FXML
    private Label chooseNodeLabel;
    @FXML
    private Button refreshAdvertisementButton;
    @FXML
    private RadioButton componentIdNodeRadioButton, anyNodeRadioButton, hardwareTypeNodeRadioButton;
    @FXML
    private ProgressBar advertisementFetchProgressBar;
    @FXML
    private VBox advertisementFetchOverlay;
    @FXML
    private Button nodesVizButton;

    @FXML
    private CheckBox onlyAvailableNodes;


    private final ObjectProperty<AdvertisementFetchManager.Resources.RefreshAdvertisementTask> activeRefreshTask =
            new SimpleObjectProperty<>();
    private final ObjectProperty<AdvertisementFetchManager.Resources> activeResources = new SimpleObjectProperty<>();

//fedmon-v3 refactor: no direct link to advertisement or FXRspecNode anymore. Use JFedGuiConfig to get
//        list of componentIds, list of hardwareTypes or list of diskimages. you don't know here where that info comes from,
//        but you can request advertisement refresh and that MIGHT change the list when completed
//    private final ListProperty<FXRspecNode> nodesList = new SimpleListProperty<>();


    //fedmon-v3 refactor: don't think we need this
//    private final FXModelRspec userSpecifiedNodesModel = FXRspecFactory.getInstance().createModelRspec("advertisement");
//    private final FXRspecNode userSpecifiedAdvNode;

    public ComponentManagerPropertiesPane(final Resource resource,
                                          FXRspecNode node, boolean initialAvailableNodesOnly,
                                          JFedGuiConfig config, AdvertisementFetchManager advertisementFetchManager,
                                          MapDialogFactory mapDialogFactory,
                                          BonfireLicenseUtil bonfireLicenseUtil) {
        assert resource != null;
        assert node != null;

        this.resource = resource;
        this.config = config;
        this.advertisementFetchManager = advertisementFetchManager;
        this.bonfireLicenseUtil = bonfireLicenseUtil;
        this.mapDialogFactory = mapDialogFactory;

        //initialize FXML
        FXMLUtil.injectFXML(this);

        onlyAvailableNodes.setSelected(initialAvailableNodesOnly);
        //the order of bindings is important!

        Server server = resource.getServer();
        assert server != null : "Resource "+resource.getId()+" has no server";

        componentIdComboBox.setItems(config.getComponentInfo(server, initialAvailableNodesOnly ? JFedGuiConfig.Scope.AVAILABLE_ONLY : JFedGuiConfig.Scope.ALL));
        hardwareTypesComboBox.setItems(config.getHardwareTypes(server, initialAvailableNodesOnly ? JFedGuiConfig.Scope.AVAILABLE_ONLY : JFedGuiConfig.Scope.ALL));

//        //bind componentId-combobox to advertisement
//        nodesList.addListener(new InvalidationListener() {
//            @Override
//            public void invalidated(Observable observable) {
//                //backup selected node
//                FXRspecNode selectedValue = componentIdComboBox.getValue();
//
//                //update list
//                componentIdComboBox.setItems(nodesList.get());
//                if (userSpecifiedAdvNode != null)
//                    componentIdComboBox.getItems().add(userSpecifiedAdvNode);
//
//                //try to restore the selected value
//                if (selectedValue != null) {
//                    assert (selectedValue.getComponentId() != null);
//                    Optional<FXRspecNode> newSelectedValue = componentIdComboBox.getItems().stream()
//                            .filter(newAdvNode -> selectedValue.getComponentId().equals(newAdvNode.getComponentId()))
//                            .findFirst();
//
//                    componentIdComboBox.setValue(newSelectedValue.orElse(null));
//                }
//            }
//        });
//
//        //bind hardwaretypes-combobox to advertisement
//        nodesList.addListener(new InvalidationListener() {
//            @Override
//            public void invalidated(Observable observable) {
//                List<HardwareType> hardwareTypes =
//                        nodesList.get().stream().flatMap(node -> node.getHardwareTypes().stream())
//                                .filter(hardwareType ->
//                                        !HARDWARE_TYPE_NAMES_SUBSTRING_BLACKLIST.stream()
//                                                .anyMatch(blacklistItem ->
//                                                        hardwareType.getName().toLowerCase()
//                                                                .contains(blacklistItem))
//                                                &&
//                                                !HARDWARE_TYPE_NAMES_BLACKLIST.contains(hardwareType.getName().toLowerCase())
//                                )
//                                .distinct()
//                                .sorted(HARDWARE_TYPE_COMPARATOR)
//                                .collect(Collectors.toList());
//
//                if (!hardwareTypes.contains(hardwareTypesComboBox.getSelectionModel().getSelectedItem()))
//                    hardwareTypesComboBox.getSelectionModel().clearSelection();
//
//
//                hardwareTypesComboBox.getItems().setAll(hardwareTypes);
//            }
//        });

        //1. bind refresh overlay to activeRefreshTask
        advertisementFetchOverlay.visibleProperty().unbind();
        advertisementFetchOverlay.visibleProperty().bind(activeRefreshTask.isNotNull());

        activeRefreshTask.addListener((observable, oldValue, newValue) -> {
            if (oldValue != null) {
                advertisementFetchProgressBar.progressProperty().unbind();
            }

            if (newValue != null) {
                advertisementFetchProgressBar.progressProperty().bind(newValue.progressProperty());
            }
        });

        //2. bind activeRefreshTask to activeResources
        activeResources.addListener(new ChangeListener<AdvertisementFetchManager.Resources>() {
            @Override
            public void changed(ObservableValue<? extends AdvertisementFetchManager.Resources> observable,
                                AdvertisementFetchManager.Resources oldValue, AdvertisementFetchManager.Resources newValue) {
                activeRefreshTask.unbind();
                activeRefreshTask.bind(newValue.currentRefreshAdvertisementTaskProperty());

                //TODO fedmon-v3: this may be removed, processing the listresources result should be done by AdvertisementFetchManager
//                nodesList.set(cmiManager.filterComponentIds(ComponentManagerPropertiesPane.this.resource, activeResources.get().getAdvertisementComponentIds()));
            }
        });

        //3. bind activeResources
        activeResources.set(initialAvailableNodesOnly ? advertisementFetchManager.getAvailableResources(server)
                : advertisementFetchManager.getAllResources(server));

        //initialize view
//        componentIdComboBox.setConverter(StringConverters.ADV_RSPECNODE_CONVERTER);
//        hardwareTypesComboBox.setConverter(StringConverters.HARDWARE_TYPE_CONVERTER);

        //initialize bindings
        componentIdComboBox.disableProperty().bind(componentIdNodeRadioButton.selectedProperty().not());
        onlyAvailableNodes.disableProperty().bind(anyNodeRadioButton.selectedProperty());

        hardwareTypesComboBox.disableProperty().bind(hardwareTypeNodeRadioButton.selectedProperty().not());

        onlyAvailableNodes.selectedProperty().addListener(observable -> {
            if (onlyAvailableNodes.isSelected()) {
                this.activeResources.setValue(advertisementFetchManager.getAvailableResources(server));
            } else {
                this.activeResources.setValue(advertisementFetchManager.getAllResources(server));
            }

            if (!this.activeResources.get().isComponentIdsAvailable())
                refreshAdvertisement();

            componentIdComboBox.setItems(config.getComponentInfo(server, initialAvailableNodesOnly ? JFedGuiConfig.Scope.AVAILABLE_ONLY : JFedGuiConfig.Scope.ALL));
            hardwareTypesComboBox.setItems(config.getHardwareTypes(server, initialAvailableNodesOnly ? JFedGuiConfig.Scope.AVAILABLE_ONLY : JFedGuiConfig.Scope.ALL));
        });

        //initialize pane with values from node
        if (node.getComponentId() != null) {
            Optional<JFedGuiConfig.ComponentInfo> componentId = componentIdComboBox.getItems().stream()
                    .filter(compInfo -> node.getComponentId().equals(compInfo.getUrn()))
                    .findFirst();

            if (componentId.isPresent()) {
                componentIdComboBox.getSelectionModel().select(componentId.get());
                componentIdNodeRadioButton.setSelected(true);

//                userSpecifiedAdvNode = null;
            } else {
//                //make a copy of the node, so we can store it for the list.
//                userSpecifiedAdvNode =
//                        FXRspecFactory.getInstance().copyNode(userSpecifiedNodesModel, node, RspecNode.InterfaceCopyMethod.NO_COPY);
//                userSpecifiedNodesModel.addNode(userSpecifiedAdvNode);

                //add it to the list of known componentIds in JFedGuiConfig
                JFedGuiConfig.ComponentInfo addedInfo = config.addUserComponentInfo(server.getId(), node.getComponentId());
                componentIdComboBox.getSelectionModel().select(addedInfo);
                componentIdNodeRadioButton.setSelected(true);
            }
        } else if (!node.getHardwareTypes().isEmpty()) {
//            userSpecifiedAdvNode = null;

            if (node.getHardwareTypes().size() > 1) {
                LOG.warn("A node can only have one hardware type in request, but we found {}. Taking first one, and ignoring the rest!", node.getHardwareTypes().size());
            }

            HardwareType userSpecifiedHardwareType = node.getHardwareTypes().get(0);

            JFedGuiConfig.HardwareTypeInfo addedHardwareTypeInfo = config.addUserHardwareTypeInfo(server.getId(), userSpecifiedHardwareType.getName(), null);
            hardwareTypesComboBox.getSelectionModel().select(addedHardwareTypeInfo);
            hardwareTypeNodeRadioButton.setSelected(true);
        } else {
//            userSpecifiedAdvNode = null;
        }

        //add this listener AFTER  the possible addition to hardwareTypesComboBox. This to prevent an unnecessary
        //advertisement call when componentId or hardwareType selected
        //TODO: fedmon-v3: check first if this is cached/known, so that advertisement is not requested when not needed
        //                 probably need to change gui to let user know data is cached, and can be refreshed to get latest
        componentIdNodeRadioButton.selectedProperty().addListener(observable -> {
            if (componentIdNodeRadioButton.isSelected() &&
                    !activeResources.get().isComponentIdsAvailable()) {
                refreshAdvertisement();
            }
        });

        hardwareTypeNodeRadioButton.selectedProperty().addListener(observable -> {
            if (hardwareTypeNodeRadioButton.isSelected() &&
                    !activeResources.get().isComponentIdsAvailable()) {
                refreshAdvertisement();
            }
        });

        //but make sure that a refresh is done if a fixed node assignment is mandatory!
        if (this.resource.getRequireFixedNodeAssignment()) {
            anyNodeRadioButton.setDisable(true);
            componentIdNodeRadioButton.setSelected(true);
        }

        if (!this.resource.getAllowFixedNodeAssignment()) {
            assert !this.resource.getRequireFixedNodeAssignment();
            componentIdNodeRadioButton.setVisible(false);
            componentIdComboBox.setVisible(false);
            nodesVizButton.setVisible(false);

            componentIdNodeRadioButton.setSelected(false);
        }

        if (!this.resource.getAllowFixedNodeAssignment() && !this.resource.getAllowHardwareType()) {
            //if both are not allowed, remove everything related to node selection

            assert !this.resource.getRequireFixedNodeAssignment();
            assert !this.resource.getRequireHardwareType();
            anyNodeRadioButton.setVisible(false);
            chooseNodeLabel.setVisible(false);
            onlyAvailableNodes.setVisible(false);
        }

        if (!this.resource.getAllowFixedNodeAssignment() &&
                !this.resource.getAllowHardwareType() &&
                !this.resource.getAllowDiskImage()) {
            //if these 3 are not allowed, remove everything related to advertisement RSpec
            refreshAdvertisementButton.setVisible(false);
        }

//        if (!this.resource.getAllowHardwareType()) {
//            hardwareTypesComboBox.setVisible(false);
//            hardwareTypeNodeRadioButton.setVisible(false);
//            hardwareTypeNodeRadioButton.setSelected(false);
//        }
        hardwareTypesComboBox.setVisible(this.resource.getAllowHardwareType());
        hardwareTypeNodeRadioButton.setVisible(this.resource.getAllowHardwareType());

        if (this.resource.getRequireHardwareType()) {
            hardwareTypeNodeRadioButton.setSelected(true); //not really wat we want?
        }



        //check if viz-button must be enabled for this testbed
        if (this.resource.getAllowFixedNodeAssignment()) {
            JFedGuiConfig.NodesVisualization nodesVisualization = config.getNodesVisualization(this.resource.getServer());
            if (nodesVisualization != null) {
                nodesVizButton.disableProperty().bind(componentIdComboBox.disableProperty());

                if (nodesVisualization == JFedGuiConfig.NodesVisualization.MAP) {
                    nodesVizButton.setOnAction(this::showNodesMap);

                    //only show maps-button if running on a 64-bit JVM (see FEDIBBTDEV-1344)
                    String arch = System.getProperty("os.arch");

                    if (arch.equals("x86")) {
                        nodesVizButton.setVisible(false);
                    }
                } else if (nodesVisualization == JFedGuiConfig.NodesVisualization.IMAGE) {
                    nodesVizButton.setOnAction(this::showNodesImageMap);
                }
            }
        }

    }


    private void showNodesImageMap(ActionEvent actionEvent) {
        ImageMap<GeniUrn> imageMap = config.getImageMap(resource.getServer());
        assert imageMap != null;
        JFedGuiConfig.ComponentInfo selectedNode =
                mapDialogFactory.showNodeImageMapDialog(this.getScene().getWindow(), imageMap,
                        config.getComponentInfo(resource.getServer(), onlyAvailableNodes.isSelected() ? JFedGuiConfig.Scope.AVAILABLE_ONLY : JFedGuiConfig.Scope.ALL));

        if (selectedNode != null) {
            componentIdComboBox.setValue(selectedNode);
        }
    }

    private void showNodesMap(ActionEvent actionEvent) {
        JFedGuiConfig.ComponentInfo selectedNode =
                mapDialogFactory.showNodeMapDialog(this.getScene().getWindow(),
                        config.getComponentInfo(resource.getServer(), onlyAvailableNodes.isSelected() ? JFedGuiConfig.Scope.AVAILABLE_ONLY : JFedGuiConfig.Scope.ALL)
                );

        if (selectedNode != null) {
            componentIdComboBox.setValue(selectedNode);
        }
    }


    private static boolean isConnectivityException(Throwable t) {

        while (t != null) {
            if (t.getCause() instanceof JFedConnectivityException)
                return true;
            else
                t = t.getCause();
        }
        return false;
    }


    private void refreshAdvertisement() {
        final AdvertisementFetchManager.Resources.RefreshAdvertisementTask updateAdvTask =
                activeResources.get().refreshAdvertisement();

        updateAdvTask.setOnFailed(workerStateEvent -> {
            //only show messages if pane is still visible
            if (ComponentManagerPropertiesPane.this.getParent() != null) {
                if (isConnectivityException(updateAdvTask.getException())) {
                    JFDialogs.create()
                            .owner(ComponentManagerPropertiesPane.this)
                            .message(String.format("A connectivity exception occurred while fetching the advertisement RSpec from %s.\n\n" +
                                            "Verify your internet connection, and consider enabling a proxy in the jFed Preferences.",
                                    resource.getServer().getName()))
                            .masthead("Fetching node advertisement failed")
                            .title("Fetching node advertisement failed")
                            .showException(updateAdvTask.getException());


                } else if(BonfireLicenseUtil.isBonfireLicenseException(updateAdvTask.getException())) {
                    bonfireLicenseUtil.showBonfireLicenseAlert(updateAdvTask.getException());

                }else {
                    JFDialogs.create()
                            .owner(ComponentManagerPropertiesPane.this)
                            .message(String.format("An error occurred while fetching the advertisement RSpec from %s.",
                                    resource.getServer().getName()))
                            .masthead("Fetching advertisement RSpec failed")
                            .title("Fetching advertisement RSpec failed")
                            .showException(updateAdvTask.getException());

                }
            }
        });
    }

    public boolean validate() {
        if (componentIdNodeRadioButton.isSelected() && componentIdComboBox.getSelectionModel().isEmpty()) {
            JFDialogs.create()
                    .owner(this)
                    .message("Please select a node from the list")
                    .masthead("No node selected")
                    .showError();
            return false;
        }

        return true;
    }

    public GeniUrn getSelectedComponentId() {
        if (!componentIdNodeRadioButton.isSelected())
            return null;

        JFedGuiConfig.ComponentInfo res = componentIdComboBox.getSelectionModel().getSelectedItem();
        return res == null ? null : res.getUrn();
    }

    public ReadOnlyBooleanProperty isSpecificNodeChosen() {
        return componentIdNodeRadioButton.selectedProperty();
    }

    public HardwareType getSelectedHardwareType() {
        if (!hardwareTypeNodeRadioButton.isSelected())
            return null;

        JFedGuiConfig.HardwareTypeInfo hardwareTypeInfo = hardwareTypesComboBox.getSelectionModel().getSelectedItem();
        String hardwareTypeName = hardwareTypeInfo == null ? null : hardwareTypeInfo.getHardwareTypeName();
        if (hardwareTypeName == null)
            return null;

        return new HardwareType(hardwareTypeName);
    }

    @FXML
    private void onRefreshAdvertisementAction() {
        refreshAdvertisement();
    }

    @FXML
    private void showTestbedMap() {
    }

    public boolean isActiveNodesOnly() {
        return onlyAvailableNodes.isSelected();
    }
}
