package be.iminds.ilabt.jfed.highlevel.util;

import be.iminds.ilabt.jfed.lowlevel.GeniUser;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnection;
import be.iminds.ilabt.jfed.lowlevel.ssh_key_info.GeniUserSshKeyInfo;
import be.iminds.ilabt.jfed.preferences.ProxyPreferencesManager;
import be.iminds.ilabt.jfed.util.SshProxySocketFactory;

import javax.annotation.Nullable;
import javax.inject.Inject;
import javax.inject.Singleton;
import javax.net.SocketFactory;
import java.io.IOException;

/**
 * User: twalcari
 * Date: 5/27/2016
 * Time: 09:59
 */
@Singleton
public class ProxySocketFactoryProviderImpl implements ProxySocketFactoryProvider {

    private final GeniUserProvider geniUserProvider;
    private final ProxyPreferencesManager proxyPreferencesManager;

    @Inject
    public ProxySocketFactoryProviderImpl(GeniUserProvider geniUserProvider,
                                          ProxyPreferencesManager proxyPreferencesManager) {
        this.geniUserProvider = geniUserProvider;
        this.proxyPreferencesManager = proxyPreferencesManager;
    }

    @Nullable
    @Override
    public SocketFactory createProxySocketFactory() throws IOException {
        if (!proxyPreferencesManager.isProxyEnabledForJfed())
            return null;
        else return createSshProxySocketFactory();
    }

    @Override
    @Nullable
    public SocketFactory createProxySocketFactoryForSsh() throws IOException {
        if (!proxyPreferencesManager.isProxyEnabledForSsh())
            return null;
        else return createSshProxySocketFactory();
    }

    private SocketFactory createSshProxySocketFactory() throws IOException {
        final GeniUser loggedInGeniUser = geniUserProvider.getLoggedInGeniUser();

        GeniUserSshKeyInfo sshKeyInfo = new GeniUserSshKeyInfo(loggedInGeniUser);

        JFedConnection.SshProxyInfo proxyInfo =
                proxyPreferencesManager.chooseUserProxy(
                        loggedInGeniUser.getUserAuthorityServer().getTestbed().getProxies(),
                        loggedInGeniUser,
                        sshKeyInfo);

        return SshProxySocketFactory.create(proxyInfo);
    }
}
