package be.iminds.ilabt.jfed.lowlevel.ssh_key_info;

import be.iminds.ilabt.jfed.lowlevel.GeniUser;
import be.iminds.ilabt.jfed.util.JFedUtils;
import be.iminds.ilabt.jfed.util.KeyUtil;
import be.iminds.ilabt.jfed.util.tmp_file_helpers.PermanentFile;
import be.iminds.ilabt.jfed.util.tmp_file_helpers.TmpContentFile;
import be.iminds.ilabt.jfed.util.tmp_file_helpers.TmpFile;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import java.io.File;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.util.Date;

/**
 * User: twalcari
 * Date: 5/27/2016
 * Time: 09:49
 */
public class GeniUserSshKeyInfo extends AbstractSshKeyInfo{
    private static final Logger LOG = LoggerFactory.getLogger(GeniUserSshKeyInfo.class);
    @Nonnull
    protected final GeniUser geniUser;

    public GeniUserSshKeyInfo(@Nonnull GeniUser geniUser) {
        this.geniUser = geniUser;
    }

    @Override
    public TmpFile getPrivateKeyFile() {
        if (geniUser.getPrivateKeyFile() != null) {
            PermanentFile res = new PermanentFile(geniUser.getPrivateKeyFile());
            ensureMatchingPubFile(res);
            return res;
        }

        LOG.debug("geniUser has no privateKeyFile. geniUser class: " + geniUser.getClass().getName());

        return getUnencryptedPrivateKeyFile();
    }

    @Override
    public TmpFile getUnencryptedPrivateKeyFile() {
        File tmpDir = new File(JFedUtils.getUserDataDirectoryFile(), "tmp");
        if (!tmpDir.exists()) tmpDir.mkdir();
        long nr = new Date().getTime();
        String pemKey = new String(KeyUtil.privateKeyToAnyPem(geniUser.getPrivateKey()));
        TmpContentFile unlockedPrivateKeyFile =
                new TmpContentFile("sshKeyUsrUE", ".pem", pemKey, true /*assureUserOnlyPermissions*/);
        ensureMatchingPubFile(unlockedPrivateKeyFile);
        return unlockedPrivateKeyFile;
    }

    @Override
    public boolean hasUnencryptedPrivateKey() {
        return true;
    }

    @Override
    public PublicKey getPublicKey() {
        return geniUser.getPublicKey();
    }

    @Override
    public PrivateKey getPrivateKey() {
        return geniUser.getPrivateKey();
    }

    @Nonnull
    public GeniUser getGeniUser() {
        return geniUser;
    }
}
