package be.iminds.ilabt.jfed.ui.javafx.scanner;

import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Server;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Testbed;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.authority.JFedTestbedInfoSource;
import be.iminds.ilabt.jfed.lowlevel.authority.StoredTestbedInfo;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.ui.javafx.dialogs.JFDialogs;
import be.iminds.ilabt.jfed.util.IOUtils;
import be.iminds.ilabt.jfed.util.JavaFXLogger;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.SerializationFeature;
import io.dropwizard.jackson.Jackson;
import javafx.fxml.FXML;
import javafx.scene.control.Button;
import javafx.scene.control.ButtonType;
import javafx.scene.control.TextArea;
import javafx.stage.FileChooser;
import javafx.stage.Stage;
import javafx.stage.Window;
import org.slf4j.LoggerFactory;

import javax.inject.Inject;
import java.io.File;
import java.util.Arrays;
import java.util.Optional;
import java.util.regex.Pattern;

/**
 * ScanResultPanelController
 */
public class ScanResultPanelController {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(ScanResultPanelController.class);

    @FXML
    private TextArea xmlTextArea;
    @FXML
    private Button addButton;

    private final JavaFXLogger logger;
    private final GeniUserProvider geniUserProvider;
    private final JFedPreferences jFedPreferences;

    private Stage stage;
    private Testbed testbed;
    private boolean scanAgain;

    @Inject
    public ScanResultPanelController(JavaFXLogger logger, GeniUserProvider geniUserProvider, JFedPreferences jFedPreferences) {
        this.logger = logger;
        this.geniUserProvider = geniUserProvider;
        this.jFedPreferences = jFedPreferences;
        this.scanAgain = false;
    }

    @FXML
    private void initialize() { }

    public void showSnippet() {
        assert xmlTextArea != null;
        if (testbed != null) {
            ObjectMapper MAPPER = Jackson.newObjectMapper();
            String testbedJson = null;
            try {
                testbedJson = MAPPER.writer().with(SerializationFeature.INDENT_OUTPUT).writeValueAsString(testbed);
            } catch (JsonProcessingException e) {
                testbedJson = "Error serializing to JSON:\n"+e.getMessage()+" ("+e.getOriginalMessage()+") @ "+e.getLocation();
            }
            xmlTextArea.setText(testbedJson);
        } else
            xmlTextArea.setText("No testbed");

        addButton.setVisible(false);
    }
    public void showStandalone() {
        assert xmlTextArea != null;
        if (testbed != null) {
            ObjectMapper MAPPER = Jackson.newObjectMapper();
            String testbedsJson = null;
            try {
                testbedsJson = MAPPER.writer().with(SerializationFeature.INDENT_OUTPUT).writeValueAsString(Arrays.asList(testbed));
            } catch (JsonProcessingException e) {
                testbedsJson = "Error serializing to JSON:\n"+e.getMessage()+" ("+e.getOriginalMessage()+") @ "+e.getLocation();
            }
            xmlTextArea.setText(testbedsJson);
        } else
            xmlTextArea.setText("No testbed");

        addButton.setVisible(true);
    }

    public void save() {
        FileChooser fc = new FileChooser();
        //set the default open dir
        if (System.getProperty("user.home") != null)
            fc.setInitialDirectory(new File(System.getProperty("user.home")));
        fc.setInitialFileName("new_authority.xml");
        File file = fc.showSaveDialog(null);

        if (file == null)
            return;

        IOUtils.stringToFile(file, xmlTextArea.getText());
    }

    private Window getOwner() {
        return addButton.getScene().getWindow();
    }

    public void add() {
        File dir = JFedTestbedInfoSource.getExtraTestbedsDir();
        Server server = testbed.getServers().get(0);
        String name = server.getUrnTld().replaceAll(Pattern.quote("."), "_");
        File file = new File(dir, name+".xml");

        if (file.exists()) {
            Optional<ButtonType> reply = JFDialogs.create()
                    .owner(getOwner())
                    .message("Overwrite existing \"" + file.getAbsolutePath() + "\"?")
                    .masthead("File exists")
                    .showConfirm();

            if (!reply.isPresent() && reply.get() != ButtonType.YES) {
                return;
            }
        }

        try {
            if (!dir.exists())
                dir.mkdir();

            IOUtils.stringToFile(file, xmlTextArea.getText());
            System.out.println("Wrote testbed to jFed: " + file.getAbsolutePath());
            LOG.debug("Wrote testbed to jFed: " + file.getAbsolutePath());
            JFDialogs.create()
                    .owner(getOwner())
                    .message("The Authority XML has been successfully stored to the jFed \"extra authorities dir\".\n" +
                            "Filename: "+file.getAbsolutePath())
                    .masthead("Successfully stored info")
                    .showInformation();
        } catch (Exception | AssertionError t) {
            LOG.error("Error writing \""+file.getAbsolutePath()+"\"", t);
            JFDialogs.create()
                    .owner(getOwner())
                    .message("Error writing \"" + file.getAbsolutePath() + "\": "+t.getMessage())
                    .masthead("Error writing file")
                    .showError();
        }
    }

    public void scan() {
        scanAgain = true;
        if (stage != null)
            stage.close();
    }
    public void close() {
        scanAgain = false;
        if (stage != null)
            stage.close();
    }

    public void setTestbed(Testbed testbed) {
        this.testbed = testbed;
    }

    public void setStage(Stage stage) {
        this.stage = stage;
    }

    public boolean isScanAgain() {
        return scanAgain;
    }
}
